@@ -93,10 +93,11 @@ void srl_decoder_destructor_hook(pTHX_ void *p);                    /* destructo
 
 /* the top level components of the decode process - called by srl_decode_into() */
 SRL_STATIC_INLINE void srl_begin_decoding(pTHX_ srl_decoder_t *dec, SV *src, UV start_offset);       /* set up the decoder to handle a given var */
-SRL_STATIC_INLINE void srl_read_header(pTHX_ srl_decoder_t *dec);                    /* validate header */
+SRL_STATIC_INLINE void srl_read_header(pTHX_ srl_decoder_t *dec, SV *header_user_data); /* read/validate header */
 SRL_STATIC_INLINE void srl_read_single_value(pTHX_ srl_decoder_t *dec, SV* into);   /* main recursive dump routine */
 SRL_STATIC_INLINE void srl_finalize_structure(pTHX_ srl_decoder_t *dec);             /* optional finalize structure logic */
 SRL_STATIC_INLINE void srl_clear_decoder(pTHX_ srl_decoder_t *dec);                 /* clean up decoder after a dump */
+SRL_STATIC_INLINE void srl_clear_decoder_body_state(pTHX_ srl_decoder_t *dec);      /* clean up after each document body */
 
 /* the internal routines to handle each kind of object we have to deserialize */
 SRL_STATIC_INLINE SV *srl_read_alias(pTHX_ srl_decoder_t *dec);
@@ -148,8 +149,31 @@ SRL_STATIC_INLINE SV *srl_read_extend(pTHX_ srl_decoder_t *dec, SV* into);
         } STMT_END
 #endif
 
+#define SRL_ASSERT_REF_PTR_TABLES(dec) STMT_START {     \
+            if (expect_false( !(dec)->ref_stashes )) {  \
+                (dec)->ref_stashes = PTABLE_new();      \
+                (dec)->ref_bless_av = PTABLE_new();     \
+            }                                           \
+        } STMT_END
 
-/* PUBLIC ROUTINES ROUTINES */
+
+STATIC void
+srl_ptable_debug_callback(PTABLE_ENTRY_t *e)
+{
+    dTHX;
+    printf("KEY=%lu\nVALUE:\n", (unsigned long)e->key);
+    sv_dump((SV *)e->value);
+    printf("\n");
+}
+
+STATIC void
+srl_ptable_debug_dump(pTHX_ PTABLE_t *tbl)
+{
+    PTABLE_debug_dump(tbl, srl_ptable_debug_callback);
+}
+
+
+/* PUBLIC ROUTINES */
 
 /* Builds the C-level configuration and state struct.
  * Automatically freed at scope boundary. */
@@ -231,17 +255,16 @@ srl_decoder_destructor_hook(pTHX_ void *p)
     }
 }
 
-/* This is the main routine to deserialize a structure.
- * It rolls up all the other "top level" routines into one
- */
-SV *
-srl_decode_into(pTHX_ srl_decoder_t *dec, SV *src, SV* into, UV start_offset)
+/* Logic shared by the various decoder entry points. */
+SRL_STATIC_INLINE void
+srl_decode_into_internal(pTHX_ srl_decoder_t *dec, SV *src, SV *header_into, SV *body_into, UV start_offset)
 {
     assert(dec != NULL);
     if (SvUTF8(src))
         sv_utf8_downgrade(src, 0);
     srl_begin_decoding(aTHX_ dec, src, start_offset);
-    srl_read_header(aTHX_ dec);
+    srl_read_header(aTHX_ dec, header_into);
+    SRL_UPDATE_BODY_POS(dec);
     if (SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_DECOMPRESS_SNAPPY)) {
         /* uncompress */
         uint32_t dest_len;
@@ -280,6 +303,7 @@ srl_decode_into(pTHX_ srl_decoder_t *dec, SV *src, SV* into, UV start_offset)
         old_pos = dec->pos;
         dec->buf_start = buf;
         dec->pos = buf + sereal_header_len;
+        SRL_UPDATE_BODY_POS(dec);
         dec->buf_end = dec->pos + dest_len;
         dec->buf_len = dest_len + sereal_header_len;
 
@@ -293,11 +317,8 @@ srl_decode_into(pTHX_ srl_decoder_t *dec, SV *src, SV* into, UV start_offset)
         }
     }
 
-    if (expect_true(!into)) {
-        into= sv_2mortal(newSV_type(SVt_NULL));
-    }
-    srl_read_single_value(aTHX_ dec, into);
-    /* assert(dec->pos == dec->buf_end); For now we disable this */
+    /* The actual document body deserialization: */
+    srl_read_single_value(aTHX_ dec, body_into);
     if (expect_false(SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_NEEDS_FINALIZE))) {
         srl_finalize_structure(aTHX_ dec);
     }
@@ -315,9 +336,44 @@ srl_decode_into(pTHX_ srl_decoder_t *dec, SV *src, SV* into, UV start_offset)
     }
 
     srl_clear_decoder(aTHX_ dec);
-    return into;
 }
 
+/* This is the main routine to deserialize just the header of a document. */
+SV *
+srl_decode_header_into(pTHX_ srl_decoder_t *dec, SV *src, SV* header_into, UV start_offset)
+{
+    assert(dec != NULL);
+    if (SvUTF8(src))
+        sv_utf8_downgrade(src, 0);
+    srl_begin_decoding(aTHX_ dec, src, start_offset);
+    if (header_into == NULL)
+        header_into = sv_newmortal();
+    srl_read_header(aTHX_ dec, header_into);
+    return header_into;
+}
+
+/* This is the main routine to deserialize a Sereal document
+ * w/o data in header. */
+SV *
+srl_decode_into(pTHX_ srl_decoder_t *dec, SV *src, SV* body_into, UV start_offset)
+{
+    if (expect_true(!body_into))
+        body_into= sv_2mortal(newSV_type(SVt_NULL));
+    srl_decode_into_internal(aTHX_ dec, src, NULL, body_into, start_offset);
+    return body_into;
+}
+
+/* This is the main routine to deserialize Sereal document body
+ * and header all at once. */
+void
+srl_decode_all_into(pTHX_ srl_decoder_t *dec, SV *src, SV *header_into, SV *body_into, UV start_offset)
+{
+    assert(header_into != NULL);
+    assert(body_into != NULL);
+    (void)srl_decode_into_internal(aTHX_ dec, src, header_into, body_into, start_offset);
+}
+
+
 /* TOP LEVEL PRIVATE ROUTINES */
 
 SRL_STATIC_INLINE void
@@ -326,8 +382,16 @@ srl_clear_decoder(pTHX_ srl_decoder_t *dec)
     if (dec->buf_start == dec->buf_end)
         return;
 
+    srl_clear_decoder_body_state(aTHX_ dec);
     SRL_DEC_RESET_VOLATILE_FLAGS(dec);
-    dec->buf_start = dec->buf_end = dec->pos = dec->save_pos = NULL;
+    dec->body_pos = dec->buf_start = dec->buf_end = dec->pos = dec->save_pos = NULL;
+}
+
+SRL_STATIC_INLINE void
+srl_clear_decoder_body_state(pTHX_ srl_decoder_t *dec)
+{
+    SRL_DEC_UNSET_OPTION(dec, SRL_F_DECODER_NEEDS_FINALIZE);
+
     if (dec->weakref_av)
         av_clear(dec->weakref_av);
 
@@ -361,23 +425,30 @@ srl_begin_decoding(pTHX_ srl_decoder_t *dec, SV *src, UV start_offset)
     dec->buf_start= dec->pos= tmp + start_offset;
     dec->buf_end= dec->buf_start + len - start_offset;
     dec->buf_len= len - start_offset;
+    SRL_SET_BODY_POS(dec, dec->buf_start);
     dec->bytes_consumed = 0;
 }
 
 SRL_STATIC_INLINE void
-srl_read_header(pTHX_ srl_decoder_t *dec)
+srl_read_header(pTHX_ srl_decoder_t *dec, SV *header_user_data)
 {
     UV header_len;
 
     /* 4 byte magic string + version/flags + hdr len at least */
     ASSERT_BUF_SPACE(dec, 4 + 1 + 1," while reading header");
     if (strnEQ((char*)dec->pos, SRL_MAGIC_STRING, 4)) {
+        unsigned int proto_version;
+
         dec->pos += 4;
         dec->proto_version_and_flags = *dec->pos++;
 
-        if (expect_false( (dec->proto_version_and_flags & SRL_PROTOCOL_VERSION_MASK) != 1 ))
+        proto_version = dec->proto_version_and_flags & SRL_PROTOCOL_VERSION_MASK;
+        if (expect_false( proto_version == 1 ))
+            SRL_DEC_SET_OPTION(dec, SRL_F_DECODER_PROTOCOL_V1); /* compat mode */
+        else if (expect_false( proto_version != 2 ))
             SRL_ERRORf1("Unsupported Sereal protocol version %u",
                     dec->proto_version_and_flags & SRL_PROTOCOL_VERSION_MASK);
+
         if ((dec->proto_version_and_flags & SRL_PROTOCOL_ENCODING_MASK) == SRL_PROTOCOL_ENCODING_RAW) {
             /* no op */
         }
@@ -396,14 +467,38 @@ srl_read_header(pTHX_ srl_decoder_t *dec)
         else
         {
             SRL_ERRORf1( "Sereal document encoded in an unknown format '%d'",
-                     (dec->proto_version_and_flags & SRL_PROTOCOL_ENCODING_MASK) >> 4 );
+                     (dec->proto_version_and_flags & SRL_PROTOCOL_ENCODING_MASK)
+                      >> SRL_PROTOCOL_VERSION_BITS);
         }
 
         /* Must do this via a temporary as it modifes dec->pos itself */
         header_len= srl_read_varint_uv_length(aTHX_ dec, " while reading header");
-        /* Skip header since we don't have any defined header-content in this
-         * protocol version. */
-        dec->pos += header_len;
+
+        if (proto_version > 1 && header_len) {
+            /* We have a protocol V2 extensible header:
+             *  - 8bit bitfield
+             *  - if lowest bit set, we have custom-header-user-data after the bitfield
+             *  => Only read header user data if an SV* was passed in to fill. */
+            const U8 bitfield = *(dec->pos++);
+            if (bitfield & SRL_PROTOCOL_HDR_USER_DATA && header_user_data != NULL) {
+                /* Do an actual document body deserialization for the user data: */
+                SRL_UPDATE_BODY_POS(dec);
+                srl_read_single_value(aTHX_ dec, header_user_data);
+                if (expect_false(SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_NEEDS_FINALIZE))) {
+                    srl_finalize_structure(aTHX_ dec);
+                }
+                srl_clear_decoder_body_state(aTHX_ dec); /* clean up for the main body decode */
+            }
+            else {
+                /* Either off in bitfield or no user data wanted, skip to end of header */
+                dec->pos += header_len - 1; /* header_len includes bitfield */
+            }
+        }
+        else {
+            /* Skip header since we don't have any defined header-content in this
+             * protocol version. */
+            dec->pos += header_len;
+        }
     } else {
         SRL_ERROR("Bad Sereal header: Does not start with Sereal magic");
     }
@@ -416,8 +511,8 @@ srl_finalize_structure(pTHX_ srl_decoder_t *dec)
     if (dec->weakref_av)
         av_clear(dec->weakref_av);
     if (dec->ref_stashes) {
-        /* FIXME the iterator may be leaked on exceptions! */
-        PTABLE_ITER_t *it = PTABLE_iter_new(dec->ref_stashes);
+        /* The iterator could be leaked on exceptions if not for PTABLE_FLAG_AUTOCLEAN. */
+        PTABLE_ITER_t *it = PTABLE_iter_new_flags(dec->ref_stashes, PTABLE_FLAG_AUTOCLEAN);
         PTABLE_ENTRY_t *ent;
 
         /* We have gotten here without error, so bless all the objects.
@@ -432,7 +527,6 @@ srl_finalize_structure(pTHX_ srl_decoder_t *dec)
                 SRL_ERROR("missing stash or ref_bless_av!");
             }
             for( len= av_len(ref_bless_av) + 1 ; len > 0 ; len-- ) {
-                /* FIXME can av_pop throw exceptions? The iterator may be leaked then... */
                 SV* obj= av_pop(ref_bless_av); /*note that av_pop does NOT refcnt dec the sv*/
                 if (SvREFCNT(obj)>1) {
                     /* It is possible that someone handcrafts a hash with a key collision,
@@ -446,7 +540,6 @@ srl_finalize_structure(pTHX_ srl_decoder_t *dec)
 #if USE_588_WORKAROUND
                         /* was blessed early, don't rebless */
 #else
-                        /* FIXME can sv_bless throw exceptions? The iterator may be leaked then... */
                         if (!nobless)
                             sv_bless(obj, stash);
 #endif
@@ -455,10 +548,8 @@ srl_finalize_structure(pTHX_ srl_decoder_t *dec)
                         SRL_ERROR("object missing from ref_bless_av array?");
                     }
                 } else {
-                    /* FIXME can warn throw exceptions? The iterator may be leaked then... */
                     warn("serialization contains a duplicated key, ignoring");
                 }
-                /* FIXME can SvREFCNT_dec throw exceptions? The iterator may be leaked then... */
                 SvREFCNT_dec(obj);
             }
         }
@@ -532,7 +623,7 @@ srl_read_varint_uv_offset(pTHX_ srl_decoder_t *dec, const char * const errstr)
 {
     UV len= srl_read_varint_uv(aTHX_ dec);
 
-    if (dec->buf_start + len >= dec->pos) {
+    if (dec->body_pos + len >= dec->pos) {
         SRL_ERRORf4("Corrupted packet%s. Offset %lu points past current position %lu in packet with length of %lu bytes long",
                 errstr, (unsigned long)len, (unsigned long)BUF_POS_OFS(dec), (unsigned long)dec->buf_len);
     }
@@ -561,11 +652,10 @@ srl_read_varint_uv_count(pTHX_ srl_decoder_t *dec, const char * const errstr)
     return len;
 }
 
-
 SRL_STATIC_INLINE void
 srl_track_sv(pTHX_ srl_decoder_t *dec, U8 *track_pos, SV *sv)
 {
-    PTABLE_store(dec->ref_seenhash, (void *)(track_pos - dec->buf_start), (void *)sv);
+    PTABLE_store(dec->ref_seenhash, (void *)(track_pos - dec->body_pos), (void *)sv);
 }
 
 
@@ -573,8 +663,10 @@ SRL_STATIC_INLINE SV *
 srl_fetch_item(pTHX_ srl_decoder_t *dec, UV item, const char const *tag_name)
 {
     SV *sv= (SV *)PTABLE_fetch(dec->ref_seenhash, (void *)item);
-    if (expect_false( !sv ))
+    if (expect_false( !sv )) {
+        /*srl_ptable_debug_dump(aTHX_ dec->ref_seenhash);*/
         SRL_ERRORf2("%s(%d) references an unknown item", tag_name, (int)item);
+    }
     return sv;
 }
 
@@ -623,7 +715,7 @@ srl_read_string(pTHX_ srl_decoder_t *dec, int is_utf8, SV* into)
     UV len= srl_read_varint_uv_length(aTHX_ dec, " while reading string");
     if (is_utf8 && SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_VALIDATE_UTF8)) {
         /* checks for invalid byte sequences. */
-        if (!is_utf8_string((U8*)dec->pos, len)) {
+        if (expect_false( !is_utf8_string((U8*)dec->pos, len) )) {
             SRL_ERROR("Invalid UTF8 byte sequence");
         }
     }
@@ -725,7 +817,7 @@ srl_read_hash(pTHX_ srl_decoder_t *dec, SV* into, U8 tag) {
     }
 
     /* Limit the maximum number of hash keys that we accept to whetever was configured */
-    if (dec->max_num_hash_entries != 0 && num_keys > dec->max_num_hash_entries) {
+    if (expect_false( dec->max_num_hash_entries != 0 && num_keys > dec->max_num_hash_entries )) {
         SRL_ERRORf2("Got input hash with %u entries, but the configured maximum is just %u",
                 (int)num_keys, (int)dec->max_num_hash_entries);
     }
@@ -746,7 +838,7 @@ srl_read_hash(pTHX_ srl_decoder_t *dec, SV* into, U8 tag) {
         U32 flags= 0;
 #endif
         ASSERT_BUF_SPACE(dec,1," while reading key tag byte for HASH");
-        tag= *dec->pos++;
+        tag= (*dec->pos++)&127;
         if (IS_SRL_HDR_SHORT_BINARY(tag)) {
             key_len= (IV)SRL_HDR_SHORT_BINARY_LEN_FROM_TAG(tag);
             ASSERT_BUF_SPACE(dec,key_len," while reading string/SHORT_BINARY key");
@@ -767,7 +859,7 @@ srl_read_hash(pTHX_ srl_decoder_t *dec, SV* into, U8 tag) {
 #endif
         } else if (tag == SRL_HDR_COPY) {
             UV ofs= srl_read_varint_uv_offset(aTHX_ dec, " while reading COPY tag");
-            from= dec->buf_start + ofs;
+            from= dec->body_pos + ofs;
             tag= *from++;
             /* note we do NOT validate these items, as we have alread read them
              * and if they were a problem we would not be here to process them! */
@@ -798,10 +890,10 @@ srl_read_hash(pTHX_ srl_decoder_t *dec, SV* into, U8 tag) {
 #else
         fetched_sv= hv_common((HV *)into, NULL, (char *)from, key_len, flags, HV_FETCH_LVALUE|HV_FETCH_JUST_SV, NULL, 0);
 #endif
-        if (!fetched_sv) {
+        if (expect_false( !fetched_sv )) {
             SRL_ERROR_PANIC(dec,"failed to hv_store");
         }
-        if ( expect_false( *dec->pos == SRL_HDR_ALIAS ) ) {
+        if (expect_false( *dec->pos == SRL_HDR_ALIAS )) {
             dec->pos++;
             *fetched_sv= srl_read_alias(aTHX_ dec);
         } else {
@@ -886,15 +978,15 @@ srl_read_objectv(pTHX_ srl_decoder_t *dec, SV* into)
     HV *stash= NULL;
 #endif
 
-    if (SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_REFUSE_OBJECTS))
+    if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_REFUSE_OBJECTS) ))
         SRL_ERROR_REFUSE_OBJECT();
 
     ofs= srl_read_varint_uv_offset(aTHX_ dec," while reading OBJECTV classname");
 
-    if ( !dec->ref_bless_av)
+    if (expect_false( !dec->ref_bless_av ))
         SRL_ERROR("Corrupted packet. OBJECTV used without preceding OBJECT to define classname");
     av= (AV *)PTABLE_fetch(dec->ref_bless_av, (void *)ofs);
-    if (NULL == av) {
+    if (expect_false( NULL == av )) {
         SRL_ERRORf1("Corrupted packet. OBJECTV references unknown classname offset: %lu", (unsigned long)ofs);
     }
     /* now deparse the thing we are going to bless */
@@ -903,7 +995,7 @@ srl_read_objectv(pTHX_ srl_decoder_t *dec, SV* into)
 #if USE_588_WORKAROUND
     /* See 'define USE_588_WORKAROUND' above for a discussion of what this does. */
     stash= PTABLE_fetch(dec->ref_stashes, (void *)ofs);
-    if (stash == NULL)
+    if (expect_false( stash == NULL ))
         SRL_ERROR("Corrupted packet. OBJECTV used without preceding OBJECT to define classname");
     if (!SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_NO_BLESS_OBJECTS))
         sv_bless(into, stash);
@@ -925,7 +1017,7 @@ srl_read_object(pTHX_ srl_decoder_t *dec, SV* into)
     U8 tag;
     U8 *from;
 
-    if (SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_REFUSE_OBJECTS))
+    if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_REFUSE_OBJECTS) ))
         SRL_ERROR_REFUSE_OBJECT();
 
     /* now find the class name - first check if this is a copy op
@@ -935,7 +1027,7 @@ srl_read_object(pTHX_ srl_decoder_t *dec, SV* into)
      * code flow in the below code */
     ASSERT_BUF_SPACE(dec,1," while reading classname tag");
 
-    storepos= BUF_POS_OFS(dec);
+    storepos= BODY_POS_OFS(dec);
     tag= *dec->pos++;
     if (IS_SRL_HDR_SHORT_BINARY(tag)) {
         key_len= SRL_HDR_SHORT_BINARY_LEN_FROM_TAG(tag);
@@ -963,7 +1055,7 @@ srl_read_object(pTHX_ srl_decoder_t *dec, SV* into)
             stash= PTABLE_fetch(dec->ref_seenhash, (void *)ofs);
         }
         if (!stash) {
-            from= dec->buf_start + ofs;
+            from= dec->body_pos + ofs;
             tag= *from++;
             /* note we do NOT validate these items, as we have alread read them
              * and if they were a problem we would not be here to process them! */
@@ -988,17 +1080,16 @@ srl_read_object(pTHX_ srl_decoder_t *dec, SV* into)
         SRL_ERROR_UNEXPECTED(dec,tag, "a class name");
     }
     if (!stash) {
-        if (expect_false( !dec->ref_stashes )) {
-            dec->ref_stashes = PTABLE_new();
-            dec->ref_bless_av = PTABLE_new();
-        }
+        SRL_ASSERT_REF_PTR_TABLES(dec);
         stash= gv_stashpvn((char *)from, key_len, flags);
         PTABLE_store(dec->ref_stashes, (void *)storepos, (void *)stash);
         av= newAV();
         sv_2mortal((SV*)av);
         PTABLE_store(dec->ref_bless_av, (void *)storepos, (void *)av);
-    } else if (NULL == (av= (AV *)PTABLE_fetch(dec->ref_bless_av, (void *)storepos)) ) {
-        SRL_ERRORf1("Panic, no ref_bless_av for %lu", (unsigned long)storepos);
+    } else {
+        SRL_ASSERT_REF_PTR_TABLES(dec);
+        if (NULL == (av= (AV *)PTABLE_fetch(dec->ref_bless_av, (void *)storepos)) )
+            SRL_ERRORf1("Panic, no ref_bless_av for %lu", (unsigned long)storepos);
     }
 
     /* we now have a stash so we /could/ bless... except that
@@ -1178,7 +1269,7 @@ srl_read_copy(pTHX_ srl_decoder_t *dec, SV* into)
         SRL_ERRORf1("COPY(%d) points out of packet", (int)item);
     }
     dec->save_pos= dec->pos;
-    dec->pos= dec->buf_start + item;
+    dec->pos= dec->body_pos + item;
     srl_read_single_value(aTHX_ dec, into);
     dec->pos= dec->save_pos;
     dec->save_pos= 0;
